import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal

import unittest
from unittest.mock import MagicMock, patch

from darkgreybox.base_model import DarkGreyModel
from darkgreybox.fit import (get_ic_params,
                             map_ic_params,
                             train_model,
                             train_models,
                             predict_model,
                             predict_models,
                             darkgreyfit,
                             reduce_results_df,
                             apply_prefit_filter)


@patch('darkgreybox.fit.apply_prefit_filter')
@patch('darkgreybox.fit.reduce_results_df')
@patch('darkgreybox.fit.predict_models')
@patch('darkgreybox.fit.train_models')
class DarkGreyFitTest(unittest.TestCase):

    def setUp(self):

        self.X_train = pd.DataFrame({
            'A0': [10, 20],
            'B': [30, 40],
            'C0': [50, 60],
            'D': [70, 80]
        })

        self.y_train = pd.Series([100, 110])
        self.Z_train = np.array([120, 130])

        self.X_test = pd.DataFrame({
            'A0': [10, 20],
            'B': [30, 40],
            'C0': [50, 60],
            'D': [70, 80]
        })

        self.y_test = pd.Series([100, 110])
        self.Z_test = np.array([120, 130])

        self.models = [MagicMock()]
        self.train_model_results = [MagicMock()]
        self.test_model_results = [MagicMock()]
        self.train_time = [1.0]
        self.test_time = [0.0]
        self.train_error = [0.9]
        self.test_error = [0.8]
        self.ic_params_map = {'param': 'value'}
        self.error_metric = MagicMock()
        self.prefit_splits = MagicMock()
        self.prefit_filter = lambda x: x < 1,

        self.train_df = pd.DataFrame({
            'model': self.models,
            'model_result': self.train_model_results,
            'start': self.X_train.index[0],
            'end': self.X_train.index[-1],
            'time': self.train_time,
            'error': self.train_error
        })

        self.test_df = pd.DataFrame({
            'model': self.models,
            'model_result': self.test_model_results,
            'start': self.X_test.index[0],
            'end': self.X_test.index[-1],
            'time': self.test_time,
            'error': self.test_error
        })

        self.columns = pd.MultiIndex.from_tuples([('train', 'model'),
                                                  ('train', 'model_result'),
                                                  ('train', 'start'),
                                                  ('train', 'end'),
                                                  ('train', 'time'),
                                                  ('train', 'error'),
                                                  ('test', 'model'),
                                                  ('test', 'model_result'),
                                                  ('test', 'start'),
                                                  ('test', 'end'),
                                                  ('test', 'time'),
                                                  ('test', 'error')])

    def test__darkgreyfit__correct_data_flow(
            self,
            mock_train_models,
            mock_predict_models,
            mock_reduce_results_df,
            mock_apply_prefit_filter):

        mock_obj_func = MagicMock()

        mock_train_models.return_value = self.train_df
        mock_predict_models.return_value = self.test_df

        expected_df = pd.DataFrame(columns=self.columns,
                                   data=[[self.models[0], self.train_model_results[0], self.X_train.index[0],
                                          self.X_train.index[-1], self.train_time[0], self.train_error[0],
                                          self.models[0], self.test_model_results[0], self.X_test.index[0],
                                          self.X_test.index[-1], self.test_time[0], self.test_error[0]]])

        actual_df = darkgreyfit(models=self.models,
                                X_train=self.X_train,
                                y_train=self.y_train,
                                X_test=self.X_test,
                                y_test=self.y_test,
                                ic_params_map=self.ic_params_map,
                                error_metric=self.error_metric,
                                prefit_splits=self.prefit_splits,
                                prefit_filter=None,
                                reduce_train_results=False,
                                method='nelder',
                                obj_func=mock_obj_func,
                                n_jobs=-1,
                                verbose=10)

        mock_train_calls = mock_train_models.call_args_list

        mock_train_calls[0] == ({'models': self.models,
                                 'X_train': self.X_train,
                                 'y_train': self.y_train,
                                 'splits': self.prefit_splits,
                                 'error_metric': self.error_metric,
                                 'method': 'nelder',
                                 'obj_func': mock_obj_func,
                                 'n_jobs': -1,
                                 'verbose': 10})

        mock_train_calls[1] == ({'models': self.models,
                                 'X_train': self.X_train,
                                 'y_train': self.y_train,
                                 'splits': None,
                                 'error_metric': self.error_metric,
                                 'method': 'nelder',
                                 'obj_func': mock_obj_func,
                                 'n_jobs': -1,
                                 'verbose': 10})

        mock_reduce_results_df.assert_not_called()

        mock_predict_models.assert_called_with(models=self.models,
                                               X_test=self.X_test,
                                               y_test=self.y_test,
                                               ic_params_map=self.ic_params_map,
                                               error_metric=self.error_metric,
                                               train_results=self.train_model_results,
                                               n_jobs=-1,
                                               verbose=10)

        assert_frame_equal(expected_df, actual_df)

    def test__darkgreyfit__with_reduce_train_results(
            self,
            mock_train_models,
            mock_predict_models,
            mock_reduce_results_df,
            mock_apply_prefit_filter):

        mock_obj_func = MagicMock()
        mock_train_models.return_value = self.train_df
        mock_predict_models.return_value = self.test_df
        mock_reduce_results_df.return_value = self.train_df

        darkgreyfit(models=self.models,
                    X_train=self.X_train,
                    y_train=self.y_train,
                    X_test=self.X_test,
                    y_test=self.y_test,
                    ic_params_map=self.ic_params_map,
                    error_metric=self.error_metric,
                    prefit_splits=self.prefit_splits,
                    prefit_filter=None,
                    reduce_train_results=True,
                    method='nelder',
                    obj_func=mock_obj_func,
                    n_jobs=-1,
                    verbose=10)

        mock_reduce_results_df.assert_called_with(self.train_df)

    def test__darkgreyfit__with_prefit_filter(
            self,
            mock_train_models,
            mock_predict_models,
            mock_reduce_results_df,
            mock_apply_prefit_filter):

        mock_obj_func = MagicMock()
        mock_train_models.return_value = self.train_df
        mock_predict_models.return_value = self.test_df
        mock_apply_prefit_filter.return_value = self.train_df

        darkgreyfit(models=self.models,
                    X_train=self.X_train,
                    y_train=self.y_train,
                    X_test=self.X_test,
                    y_test=self.y_test,
                    ic_params_map=self.ic_params_map,
                    error_metric=self.error_metric,
                    prefit_splits=self.prefit_splits,
                    prefit_filter=self.prefit_filter,
                    reduce_train_results=False,
                    method='nelder',
                    obj_func=mock_obj_func,
                    n_jobs=-1,
                    verbose=10)

        mock_apply_prefit_filter.assert_called_with(self.train_df, self.prefit_filter)


class FitTest(unittest.TestCase):

    def setUp(self):

        self.X_train = pd.DataFrame({
            'A0': [10, 20],
            'B': [30, 40],
            'C0': [50, 60],
            'D': [70, 80]
        })

        self.y_train = pd.Series([100, 110])
        self.Z_train = np.array([120, 130])

        self.X_test = pd.DataFrame({
            'A0': [10, 20],
            'B': [30, 40],
            'C0': [50, 60],
            'D': [70, 80]
        })

        self.y_test = pd.Series([100, 110])
        self.Z_test = np.array([120, 130])

    @patch('darkgreybox.fit.train_model')
    def test__train_models__not_parallel_splits_none(self, mock_train_model):

        mock_train_model.side_effect = self.mock_train_model_side_effect

        models = [MagicMock(), MagicMock()]
        error_metric = MagicMock()

        expected_df = pd.DataFrame({
            'start_date': [self.X_train.index[0]] * 2,
            'end_date': [self.X_train.index[-1]] * 2,
            'model': models,
            'model_result': ['model_result'] * 2,
            'time': [0.0] * 2,
            'method': ['nelder'] * 2,
            'error': [0.0] * 2
        })

        actual_df = train_models(models=models,
                                 X_train=self.X_train,
                                 y_train=self.y_train,
                                 error_metric=error_metric,
                                 splits=None,
                                 method='nelder',
                                 n_jobs=1,
                                 verbose=10)

        assert_frame_equal(expected_df, actual_df)

    @patch('darkgreybox.fit.train_model')
    def test__train_models__not_parallel(self, mock_train_model):

        mock_train_model.side_effect = self.mock_train_model_side_effect

        models = [MagicMock()]
        error_metric = MagicMock()

        splits = [(None, [0]), (None, [1])]

        expected_df = pd.DataFrame({
            'start_date': [self.X_train.index[0], self.X_train.index[1]],
            'end_date': [self.X_train.index[0], self.X_train.index[1]],
            'model': models * 2,
            'model_result': ['model_result'] * 2,
            'time': [0.0] * 2,
            'method': ['nelder'] * 2,
            'error': [0.0] * 2
        })

        actual_df = train_models(models=models,
                                 X_train=self.X_train,
                                 y_train=self.y_train,
                                 error_metric=error_metric,
                                 splits=splits,
                                 method='nelder',
                                 n_jobs=1,
                                 verbose=10)

        assert_frame_equal(expected_df, actual_df)

    @patch('darkgreybox.fit.train_model')
    def test__train_models__parallel(self, mock_train_model):
        # TODO
        pass

    @patch('darkgreybox.fit.copy.deepcopy')
    @patch('darkgreybox.fit.timer')
    @patch('darkgreybox.fit.get_ic_params')
    def test__train_model(self, mock_get_ic_params, mock_timer, mock_deepcopy):

        model = MagicMock()
        mock_deepcopy.return_value = model
        model_fit = MagicMock()
        model.fit.return_value = model_fit
        model_result = MagicMock()
        model_fit.predict.return_value = model_result
        model_result.Z.return_value = self.Z_train
        error_metric = MagicMock()
        error_metric.return_value = 0.95
        mock_get_ic_params.return_value = {'A0': 1}
        mock_timer.return_value = 1.0
        mock_obj_func = MagicMock()

        expected_df = pd.DataFrame({
            'start_date': [self.X_train.index[0]],
            'end_date': [self.X_train.index[-1]],
            'model': [model_fit],
            'model_result': [model_result],
            'time': [0.0],
            'method': ["splendid"],
            'error': [0.95]
        })

        actual_df = train_model(base_model=model,
                                X_train=self.X_train,
                                y_train=self.y_train,
                                error_metric=error_metric,
                                method="splendid",
                                obj_func=mock_obj_func)

        model.fit.assert_called_with(X=self.X_train.to_dict(orient='list'),
                                     y=self.y_train.values,
                                     method="splendid",
                                     ic_params={'A0': 1},
                                     obj_func=mock_obj_func)

        model_fit.predict.assert_called_with(self.X_train)
        # error_metric.assert_called_with(self.y_train.values, self.Z_train)

        assert_frame_equal(expected_df, actual_df)

    @patch('darkgreybox.fit.copy.deepcopy')
    @patch('darkgreybox.fit.timer')
    @patch('darkgreybox.fit.get_ic_params')
    def test__train_model__fit_exception(self, mock_get_ic_params, mock_timer, mock_deepcopy):

        model = MagicMock()
        mock_deepcopy.return_value = model
        model.fit.side_effect = MagicMock(side_effect=ValueError('Test'))
        mock_get_ic_params.return_value = {'A0': 1}
        error_metric = MagicMock()
        mock_timer.return_value = 1.0
        mock_obj_func = MagicMock()

        expected_df = pd.DataFrame({
            'start_date': [self.X_train.index[0]],
            'end_date': [self.X_train.index[-1]],
            'model': [np.NaN],
            'model_result': [np.NaN],
            'time': [0.0],
            'method': ["splendid"],
            'error': [np.NaN]
        })

        actual_df = train_model(base_model=model,
                                X_train=self.X_train,
                                y_train=self.y_train,
                                error_metric=error_metric,
                                method="splendid",
                                obj_func=mock_obj_func)

        model.fit.assert_called_with(X=self.X_train.to_dict(orient='list'),
                                     y=self.y_train.values,
                                     method="splendid",
                                     ic_params={'A0': 1},
                                     obj_func=mock_obj_func)

        assert_frame_equal(expected_df, actual_df)

    @patch('darkgreybox.fit.predict_model')
    def test__predict_models__not_parallel(self, mock_predict_model):

        mock_predict_model.side_effect = self.mock_predict_model_side_effect

        models = [MagicMock()] * 2
        train_results = [MagicMock()] * 2
        error_metric = MagicMock()
        ic_params_map = {}

        expected_df = pd.DataFrame({
            'start_date': [self.X_test.index[0], self.X_test.index[0]],
            'end_date': [self.X_test.index[-1], self.X_test.index[-1]],
            'model': models,
            'model_result': ['model_result'] * 2,
            'time': [0.0] * 2,
            'error': [0.0] * 2
        })

        actual_df = predict_models(models=models,
                                   X_test=self.X_test,
                                   y_test=self.y_test,
                                   ic_params_map=ic_params_map,
                                   error_metric=error_metric,
                                   train_results=train_results,
                                   n_jobs=1,
                                   verbose=10)

        mock_predict_model.assert_called_with(models[0], self.X_test, self.y_test,
                                              ic_params_map, error_metric, train_results[0])

        assert_frame_equal(expected_df, actual_df, check_dtype=False)

    @patch('darkgreybox.fit.predict_model')
    def test__predict_models__parallel(self, mock_predict_model):
        # TODO
        pass

    @patch('darkgreybox.fit.map_ic_params')
    @patch('darkgreybox.fit.timer')
    @patch('darkgreybox.base_model.DarkGreyModel.predict')
    def test__predict_model(self, mock_predict, mock_timer, mock_map_ic_params):

        model = DarkGreyModel({}, 1)
        model_result = MagicMock()
        train_result = MagicMock()
        mock_predict.return_value = model_result
        mock_timer.return_value = 1.0
        mock_map_ic_params.return_value = {'A0': 1}
        error_metric = MagicMock()
        error_metric.return_value = 0.95

        expected_df = pd.DataFrame({
            'start_date': [self.X_test.index[0]],
            'end_date': [self.X_test.index[-1]],
            'model': [model],
            'model_result': [model_result],
            'time': [0.0],
            'error': [0.95]
        })

        actual_df = predict_model(model=model,
                                  X_test=self.X_test,
                                  y_test=self.y_test,
                                  ic_params_map={'B': 1},
                                  error_metric=error_metric,
                                  train_result=train_result)

        mock_map_ic_params.assert_called_with({'B': 1}, model, self.X_test, self.y_test, train_result)

        mock_predict.assert_called_with(X=self.X_test.to_dict(orient='list'),
                                        ic_params={'A0': 1})

        assert_frame_equal(expected_df, actual_df)

    @patch('darkgreybox.fit.timer')
    def test__predict_model__not_model_instance(self, mock_timer):

        model = 1.0
        mock_timer.return_value = 1.0
        error_metric = MagicMock()

        expected_df = pd.DataFrame({
            'start_date': [self.X_test.index[0]],
            'end_date': [self.X_test.index[-1]],
            'model': [np.nan],
            'model_result': [np.nan],
            'time': [0.0],
            'error': [np.nan]
        })

        actual_df = predict_model(model=model,
                                  X_test=self.X_test,
                                  y_test=self.y_test,
                                  ic_params_map={},
                                  error_metric=error_metric,
                                  train_result=None)

        assert_frame_equal(expected_df, actual_df)

    def test__reduce_results_df(self):

        df = pd.DataFrame(data={
            'value': [0, 0, 0, 10, 20, 30, 40, 50],
            'error': [np.nan, -np.inf, np.inf, 2.0000011, 2.0000012, 2.000002, 1, 3],
            'time': [0, 0, 0, 2, 1, 3, 4, 5]
        })

        expected_df = pd.DataFrame(data={
            'value': [40, 20, 30, 50],
            'error': [1, 2.0000012, 2.000002, 3],
            'time': [4, 1, 3, 5]
        })

        actual_df = reduce_results_df(df)

        assert_frame_equal(expected_df, actual_df)

    def test__apply_prefit_filter(self):

        df = pd.DataFrame(data={
            'value': [0, 0, 0, 10, 20, 30, 40, 50],
            'error': [np.nan, -np.inf, np.inf, 2.0000011, 2.0000012, 2.000002, 1, 3],
            'time': [0, 0, 0, 2, 1, 3, 4, 5]
        })

        expected_df = pd.DataFrame(data={
            'value': [10, 40],
            'error': [2.0000011, 1],
            'time': [2, 4]
        })

        actual_df = apply_prefit_filter(df, prefit_filter=lambda x: abs(x) < 2.0000012)

        assert_frame_equal(expected_df, actual_df)

    def test__map_ic_params(self):

        train_result = [57]

        ic_params_map = {
            'A0': lambda X_test, y_test, train_result: y_test.iloc[0],
            'B0': lambda X_test, y_test, train_result: y_test.iloc[0],
            'C0': lambda X_test, y_test, train_result: X_test['C0'].iloc[0],
            'D': lambda X_test, y_test, train_result: train_result[0],
        }

        model = MagicMock()
        model.params = {'A0': 0, 'B': 1, 'C0': 2, 'D': 3}

        expected = {'A0': 100, 'C0': 50, 'D': 57}
        actual = map_ic_params(ic_params_map, model, self.X_test, self.y_test, train_result)

        self.assertEqual(expected, actual)

    def test__get_ic_params(self):

        model = MagicMock()
        model.params = {'A0': 0, 'B': 1, 'C0': 2, 'D': 3}

        expected = {'A0': 10, 'C0': 50}
        actual = get_ic_params(model, self.X_train)

        self.assertEqual(expected, actual)

    def mock_train_model_side_effect(self, base_model, X_train, y_train, error_metric, method, obj_func):
        return pd.DataFrame({
            'start_date': [X_train.index[0]],
            'end_date': [X_train.index[-1]],
            'model': [base_model],
            'model_result': ['model_result'],
            'time': [0.0],
            'method': [method],
            'error': [0.0]
        })

    def mock_predict_model_side_effect(self, model, X_test, y_test, ic_params_map, error_metric, train_result):
        return pd.DataFrame({
            'start_date': [X_test.index[0]],
            'end_date': [X_test.index[-1]],
            'model': [model],
            'model_result': ['model_result'],
            'time': [0.0],
            'error': [0.0]
        })



import numpy as np
import pandas as pd
import copy
from timeit import default_timer as timer
from joblib import Parallel, delayed

from darkgreybox import logger
from darkgreybox.base_model import DarkGreyModel


def darkgreyfit(models, X_train, y_train, X_test, y_test, ic_params_map, error_metric,
                prefit_splits=None, prefit_filter=None, reduce_train_results=True,
                method='nelder', obj_func=None, n_jobs=-1, verbose=10):
    """
    Given a list of `models` applies a prefit according to `prefit_splits`, then fits them
    to the training data and evaluates them on the test data.

    Params:
        models: list of `model.DarkGreyModel` objects
            list of models to be trained
        X_train: `pandas.DataFrame`
            A pandas DataFrame of the training input data X
        y_train: `pandas.Series`
            A pandas Series of the training input data y
        X_test: `pandas.DataFrame`
            A pandas DataFrame of the test input data X
        y_test: `pandas.Series`
            A pandas Series of the test input data y
        ic_params_map: dict
            A dictionary of mapping functions that return the
            initial condition parameters for the test set
        error_metric: function
            An error metric function that confirms to the `sklearn.metrics` interface
        prefit_splits: list
            A list of training data indices specifying sub-sections of `X_train` and `y_train`
            for the prefitting of models
        prefit_filter: function
            A function acting as a filter based on the 'error' values of the trained models
        reduce_train_results: bool
            If set to True, the training dataframe will be reduced / cleaned
            by removing nan and duplicate records
        method : str
            Name of the fitting method to use. Valid values are described in:
            `lmfit.minimize`
        obj_func: function
            The objective function to minimise during the fitting
        n_jobs: int
            The number of parallel jobs to be run as described by `joblib.Parallel`
        verbose: int
            The degree of verbosity as described by `joblib.Parallel`

    Returns:
        `pandas.DataFrame` with a record for each model's potentially viable results

    """

    prefit_df = train_models(models=models,
                             X_train=X_train,
                             y_train=y_train,
                             splits=prefit_splits,
                             error_metric=error_metric,
                             method=method,
                             obj_func=obj_func,
                             n_jobs=n_jobs,
                             verbose=verbose)

    if prefit_filter is not None:
        prefit_df = apply_prefit_filter(prefit_df, prefit_filter)

    train_df = train_models(models=prefit_df['model'].tolist(),
                            X_train=X_train,
                            y_train=y_train,
                            splits=None,
                            error_metric=error_metric,
                            method=method,
                            obj_func=obj_func,
                            n_jobs=n_jobs,
                            verbose=verbose)

    if reduce_train_results:
        train_df = reduce_results_df(train_df)

    test_df = predict_models(models=train_df['model'].tolist(),
                             X_test=X_test,
                             y_test=y_test,
                             ic_params_map=ic_params_map,
                             error_metric=error_metric,
                             train_results=train_df['model_result'].tolist(),
                             n_jobs=n_jobs,
                             verbose=verbose)

    return pd.concat([train_df, test_df], keys=['train', 'test'], axis=1)


def train_models(models, X_train, y_train, error_metric,
                 splits=None, method='nelder', obj_func=None, n_jobs=-1, verbose=10):
    """
    Trains the `models` for the given `X_train` and `y_train` training data
    for `splits` using `method`.

    Params:
        models: list of `model.DarkGreyModel` objects
            list of models to be trained
        X_train: `pandas.DataFrame`
            A pandas DataFrame of the training input data X
        y_train: `pandas.Series`
            A pandas Series of the training input data y
        error_metric: function
            An error metric function that confirms to the `sklearn.metrics` interface
        splits: list
            A list of training data indices specifying sub-sections of `X_train` and `y_train`
            for the models to be trained on
        method : str
            Name of the fitting method to use. Valid values are described in:
            `lmfit.minimize`
        obj_func: function
            The objective function to minimise during the fitting
        n_jobs: int
            The number of parallel jobs to be run as described by `joblib.Parallel`
        verbose: int
            The degree of verbosity as described by `joblib.Parallel`

    Returns:
        `pandas.DataFrame` with a record for each model's result for each split

    Example:
    ~~~~

    from sklearn.metrics import mean_squared_error
    from sklearn.model_selection import KFold

    from darkgreybox.model import TiTe
    from darkgreybox.fit import train_models


    prefit_df = train_models(models=[TiTe(train_params, rec_duration=1)],
                             X_train=X_train,
                             y_train=y_train,
                             splits=KFold(n_splits=int(len(X_train) / 24), shuffle=False).split(X_train),
                             error_metric=mean_squared_error,
                             method='nelder',
                             n_jobs=-1,
                             verbose=10)
    ~~~~
    """

    logger.info('Training models...')

    if n_jobs != 1:
        with Parallel(n_jobs=n_jobs, verbose=verbose) as p:
            df = pd.concat(p(delayed(train_model)
                             (model, X_train.iloc[idx], y_train.iloc[idx], error_metric, method, obj_func)
                             for _, idx in splits or [(None, range(len(X_train)))] for model in models),
                           ignore_index=True)

    else:
        df = pd.concat([train_model(model, X_train.iloc[idx], y_train.iloc[idx], error_metric, method, obj_func)
                        for _, idx in splits or [(None, range(len(X_train)))] for model in models],
                       ignore_index=True)

    return df


def train_model(base_model, X_train, y_train, error_metric, method='nelder', obj_func=None):
    """
    Trains a copy of `basemodel` for the given `X_train` and `y_train` training data
    using `method`.

    Params:
        base_model: `model.DarkGreyModel`
            model to be trained (a copy will be made)
        X_train: `pandas.DataFrame`
            A pandas DataFrame of the training input data X
        y_train: `pandas.Series`
            A pandas Series of the training input data y
        error_metric: function
            An error metric function that confirms to the `sklearn.metrics` interface
        method : str
            Name of the fitting method to use. Valid values are described in:
            `lmfit.minimize`
        obj_func: function
            The objective function to minimise during the fitting

    Returns:
        `pandas.DataFrame` with a single record for the fit model's result
    """

    start = timer()
    model = copy.deepcopy(base_model)

    try:
        model = model.fit(X=X_train.to_dict(orient='list'),
                          y=y_train.values,
                          method=method,
                          ic_params=get_ic_params(model, X_train),
                          obj_func=obj_func)
    except ValueError:
        end = timer()
        return pd.DataFrame({'start_date': [X_train.index[0]],
                             'end_date': [X_train.index[-1]],
                             'model': [np.NaN],
                             'model_result': [np.NaN],
                             'time': [end - start],
                             'method': [method],
                             'error': [np.NaN]})

    model_result = model.predict(X_train)
    end = timer()

    return pd.DataFrame({'start_date': [X_train.index[0]],
                         'end_date': [X_train.index[-1]],
                         'model': [model],
                         'model_result': [model_result],
                         'time': [end - start],
                         'method': [method],
                         'error': [error_metric(y_train.values, model_result.Z)]})


def predict_models(models, X_test, y_test, ic_params_map, error_metric, train_results,
                   n_jobs=-1, verbose=10):
    """
    Generates the predictions for the `models` for the given `X_test` and `y_test` test data

    Params:
        models: list of `model.DarkGreyModel` objects
            list of models to be used for the predictions
        X_test: `pandas.DataFrame`
            A pandas DataFrame of the test input data X
        y_test: `pandas.Series`
            A pandas Series of the test input data y
        ic_params_map: dict
            A dictionary of mapping functions that return the initial condition parameters
        error_metric: function
            An error metric function that confirms to the `sklearn.metrics` interface
        train_results: list of `model.DarkGreyModelResult`
            The model results of the previously trained models
        n_jobs: int
            The number of parallel jobs to be run as described by `joblib.Parallel`
        verbose: int
            The degree of verbosity as described by `joblib.Parallel`

    Returns:
        `pandas.DataFrame` with a record for each model's predictions

    Example:
    ~~~~

    from sklearn.metrics import mean_squared_error

    from darkgreybox.fit import test_models


    prefit_df = train_models(models=[trained_model_1, trained_model_2],
                             X_test=X_test,
                             y_test=y_test,
                             ic_params_map={}
                             error_metric=mean_squared_error,
                             train_results=[trained_model_result_1, trained_model_result_2],
                             n_jobs=-1,
                             verbose=10)
    ~~~~
    """

    num_models = len(models)
    logger.info(f'Generating predictions for {num_models} models...')

    if n_jobs != 1:
        with Parallel(n_jobs=n_jobs, verbose=verbose) as p:
            df = pd.concat(p(delayed(predict_model)(model, X_test, y_test, ic_params_map, error_metric, train_result)
                             for model, train_result in zip(models, train_results)), ignore_index=True)

    else:
        df = pd.concat([predict_model(model, X_test, y_test, ic_params_map, error_metric, train_result)
                        for model, train_result in zip(models, train_results)], ignore_index=True)

    return df


def predict_model(model, X_test, y_test, ic_params_map, error_metric, train_result):
    """
    Calculates redictions  of `model` for the given `X_test` and `y_test` test data.

    Params:
        model: `model.DarkGreyModel`
            model used for the prediction
        X_test: `pandas.DataFrame`
            A pandas DataFrame of the test input data X
        y_test: `pandas.Series`
            A pandas Series of the test input data y
        ic_params_map: dict
            A dictionary of mapping functions that return the initial condition parameters
        error_metric: function
            An error metric function that confirms to the `sklearn.metrics` interface
        train_result: `model.DarkGreyModelResult`
            The model result of a previously trained model

    Returns:
        `pandas.DataFrame` with a single record for the fit model's predictions
    """

    start = timer()

    if isinstance(model, DarkGreyModel):

        ic_params = map_ic_params(ic_params_map, model, X_test, y_test, train_result)

        model_result = model.predict(X=X_test.to_dict(orient='list'),
                                     ic_params=ic_params)

        end = timer()

        return pd.DataFrame({
            'start_date': [X_test.index[0]],
            'end_date': [X_test.index[-1]],
            'model': [model],
            'model_result': [model_result],
            'time': [end - start],
            'error': [error_metric(y_test.values, model_result.Z)]
        })

    else:
        end = timer()
        return pd.DataFrame({
            'start_date': [X_test.index[0]],
            'end_date': [X_test.index[-1]],
            'model': [np.NaN],
            'model_result': [np.NaN],
            'time': [end - start],
            'error': [np.NaN]
        })


def reduce_results_df(df, decimals=6):
    """
    Reduces `df` dataframe by removing nan and duplicate records

    Params:
        df: `pandas.DataFrame`
            The dataframe to be reduced / cleaned
        decimal: int
            The number of decimal points for the float comparison when removing duplicates

    Returns :
        the reduced / cleaned `pandas.DataFrame`
    """

    return (df.replace([-np.inf, np.inf], np.nan)
              .dropna()
              .round({'error': decimals})
              .sort_values('time')
              .drop_duplicates(subset=['error'], keep='first')
              .sort_values('error')
              .reset_index(drop=True))


def apply_prefit_filter(prefit_df, prefit_filter):
    """
    Applies the prefit filter to the prefit dataframe
    """
    return prefit_df[prefit_filter(prefit_df['error'])].reset_index(drop=True)


def map_ic_params(ic_params_map, model, X_test, y_test, train_result):
    """
    Maps the test initial condition parameters according to `ic_params_map`

    Parameters:
        ic_params_map: dict
            A dictionary of mapping functions that return the initial condition parameters
        model: `model.DarkGreyModel`
            model used for the prediction
        X_test: `pandas.DataFrame`
            A pandas DataFrame of the test input data X
        y_test: `pandas.Series`
            A pandas Series of the test input data y
        train_results: `model.DarkGreyModelResult`
            model result object for training data

    Returns the initial conditions parameters dict

    ~~~~
    Assuming y_test holds the internal temperatures `Ti`

    ic_params_map = {
        'Ti0': lambda X_test, y_test, train_result: y_test.iloc[0],
        'Th0': lambda X_test, y_test, train_result: y_test.iloc[0],
        'Te0': lambda X_test, y_test, train_result: train_result.Te[-1],
    }

    will map the first internal temperature in the test set to both `Ti0` and `Th0`
    and the last `Te` value from the training results to `Te0`
    ~~~~
    """

    ic_params = {}

    for key in ic_params_map:
        if key in model.params:
            ic_params[key] = ic_params_map[key](X_test, y_test, train_result)

    return ic_params


def get_ic_params(model, X_train):
    """
    Returns the initial condition parameters of a model from the training data

    Params:
        model: `model.DarkGreyModel`
            model to get initial condition parameters from
        X_train: `pandas.DataFrame`
            A pandas DataFrame of the training input data X

    Returns:
        A dictionary containing the initial conditions and their corresponding values
        as defined by the training data

    """

    # TODO: this is horrible - make this clearer and more robust
    ic_params = {}
    for key in model.params:
        if '0' in key:
            ic_params[key] = X_train.iloc[0][key]

    return ic_params



import logging.config


def enable_logging(level='INFO'):
    '''Enables global logging
    Parameters:
    level: string
        Sets the level of global logging.
        One of 'CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'
    '''

    logging_config = dict(
        version=1,
        disable_existing_loggers=False,
        formatters={
            'f': {'format':
                  '%(asctime)s %(name)-12s %(levelname)-8s %(message)s'}},
        handlers={
            'sh': {'class': 'logging.StreamHandler',
                   'formatter': 'f',
                   'stream': 'ext://sys.stdout'
                   }},
        root={
            'handlers': ['sh'],
            'level': logging.getLevelName(level)
            },
    )
    logging.config.dictConfig(logging_config)

    logger.info("Logging enabled...")


# Create shared logger object
logger = logging.getLogger(__name__)
enable_logging()



from setuptools import setup
from os import path


__VERSION__ = '0.2.0'

with open("README.md", "r") as f:
    long_description = f.read()

here = path.abspath(path.dirname(__file__))


setup(
    name='darkgreybox',
    version=__VERSION__,
    description='DarkGreyBox: An Open-Source Data-Driven Python Building Thermal Model'
                'Inspired By Genetic Algorithms and Machine Learning',
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords='python model thermal machine-learning genetic-algorithm data-science',

    # The project's main homepage.
    url='https://github.com/czagoni/darkgreybox',

    # Author details
    author='csaba zagoni',
    author_email='czagoni@greenpeace.org',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)'
    ],

    # What does your project relate to?

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=['darkgreybox'],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'lmfit~=1.0.2',
        'pandas~=1.2.3',
        'joblib~=1.0.1'
    ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': [
            'statsmodels~=0.12.2',
            'numdifftools~=0.9.39',
            'scikit-learn~=0.24.1',
            'matplotlib~=3.4.0',
            'jupyter==1.0.0',
            'notebook==6.1.5'
        ],
        'test': [
            'flake8~=3.9.0',
            'pytest~=6.2.2',
            'pytest-mock~=3.5.1',
            'pytest-cov~=2.11.1',
        ],
    },

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
    },

    include_package_data=True
)
